-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}

M.type = "auxilliary"
M.relevantDevice = rearLeftMotor, rearRightMotor, frontLeftMotor, frontRightMotor;

M.throttle_RL = 0
M.throttle_RR = 0
M.throttle_FL = 0
M.throttle_FR = 0

-- caching these because it's easier to read
-- also a stored reference has got to be faster than a string search
local motor_RL = powertrain.getDevice("rearLeftMotor")
local motor_RR = powertrain.getDevice("rearRightMotor")
local motor_FL = powertrain.getDevice("frontLeftMotor")
local motor_FR = powertrain.getDevice("frontRightMotor")

local wheel_RL = nil
local wheel_RR = nil
local wheel_FL = nil
local wheel_FR = nil


local skewStiffnessFront = 142000
local skewStiffnessRear = 177000
local distanceCOGFrontAxle = 0
local distanceCOGRearAxle = 0

local characteristicSpeed = 1
local invSquaredCharacteristicSpeed = 0 --pre calculated, used for desired yaw calculation

local configData = nil
local mass = 0

local wheelBase = 0
local invWheelBase = 0

local yawSmooth = nil
local speed = 0
local desiredYawRateSteering = 0

--shorter functions for increased performance
local abs = math.abs
local min = math.min
local max = math.max
local sqrt = math.sqrt
local acos = math.acos
local clamp = clamp
local pi = math.pi


local function printMembers(o)
	
	for key,value in pairs(o) do
		print(key .. " " .. value);
	end
	
end

-- yer basic PID settings
local yaw_P = 1
local yaw_I = 0
local yaw_D = 0


local crossAxleDeltaFactor = 1
local interAxleScaleFactor = 1
local interAxleScaleTaperPoint = 5


-- yer basic PID object
local yaw_PID = {P = yaw_P, I = yaw_I, D = yaw_D, SP = 0, MPV = 0, MPV_prev = 0, I_sum = 0}


-- yer basic PID loop
local function update_PID(pid, sp, mpv, dt)
	
	pid.MPV_prev = pid.MPV
	
	pid.MPV = mpv
	
	pid.SP = sp
	
	local D_rate = (pid.MPV - pid.MPV_prev)/dt
	
	local P_error = pid.SP - pid.MPV
	
	pid.I_sum = pid.I_sum + P_error * dt
	
	return pid.P * P_error + pid.I * pid.I_sum - pid.D * D_rate
	
end


local function updateWheelsIntermediate(dt)
	
	if electrics.values.gear == "D" then
	
		
		speed = electrics.values.wheelspeed
		local throttleInput = electrics.values.throttle_input

		-- desired yawrate will be based on wheelbase and speed

		local steeringInput = input.steering
		local wheelFront = steeringInput < 0 and wheel_FL or wheel_FR
		local wheelRear = steeringInput < 0 and wheel_RL or wheel_RR
		
		
		
		-- stock ESC steering magic
		wheelAngleFront = acos(obj:nodeVecPlanarCosRightForward(wheelFront.node1, wheelFront.node2))
		wheelAngleRear = acos(obj:nodeVecPlanarCosRightForward(wheelRear.node1, wheelRear.node2))
		if wheelAngleFront > 1.5708 then
		wheelAngleFront = (pi - wheelAngleFront)
		end
		wheelAngleFront = wheelAngleFront * fsign(-steeringInput)
		if wheelAngleFront ~= wheelAngleFront then
		wheelAngleFront = 0
		end

		if wheelAngleRear > 1.5708 then
		wheelAngleRear = (pi - wheelAngleRear)
		end
		wheelAngleRear = wheelAngleRear * fsign(-steeringInput)
		if wheelAngleRear ~= wheelAngleRear then
		wheelAngleRear = 0
		end

		M.wheelAngleFront = wheelAngleFront
		M.wheelAngleRear = wheelAngleRear
		
		
		yawRate = yawSmooth:get(obj:getYawAngularVelocity())
		
		--calculate expected yaw rate based on steering angle
		desiredYawRateSteering = -((wheelAngleFront * invWheelBase) * (speed/(1 + (speed * speed * invSquaredCharacteristicSpeed))))
		
		-- adjust the PIDSteer based on speed
		-- to get rid of juddering at lower speeds
		-- and improve response at high speeds
		local PIDSpeedFactor = (speed / characteristicSpeed)
		
		local PIDSteer = update_PID(yaw_PID, desiredYawRateSteering, yawRate, dt) * PIDSpeedFactor
		
		
		-- uncomment the below for "instant reverse" when depressing the clutch!
		-- throttleInput = throttleInput - electrics.values.clutch_input
		
		-- if throttleInput < 0 then
			-- steeringInput = -steeringInput
		-- end
		
		--
		
		
		-- apply PIDSteer to throttle values - this is the actual vectoring
		local crossAxleDelta = PIDSteer * crossAxleDeltaFactor
		
		-- inter axle scaling is based entirely on lateral g-force and the desired steering results
		-- no need to look at steering angle or understeer/oversteer which are misleading names to say the least
		-- this is a scale because just putting a negative or positive throttle won't necessarily
		-- improve traction, we need to reduce or boost the power overall regardless of direction
		local xGForce = sensors.gx2
		local interAxleScale = 1
		
		local interAxleScaleTaper = 1
		
		-- taper this influence if there is little lateral gforce
		if xGForce < interAxleScaleTaperPoint then
			interAxleScaleTaper = abs(xGForce) / interAxleScaleTaperPoint
		end
		
		if xGForce * PIDSteer > 0 then
			interAxleScale = 1 + abs(PIDSteer) * interAxleScaleFactor
		else
			interAxleScale = 1 / (1 + abs(PIDSteer) * interAxleScaleFactor)
		end
		print(interAxleScale)
		
		-- apply steering factors
		local throttle_result_FL = (throttleInput + crossAxleDelta) * interAxleScale
		local throttle_result_FR = (throttleInput - crossAxleDelta) * interAxleScale
		local throttle_result_RL = (throttleInput + crossAxleDelta) / interAxleScale
		local throttle_result_RR = (throttleInput - crossAxleDelta) / interAxleScale
		
		
		-- normalise throttle inputs in case any of them are beyond motor limits
		-- could shape this to make more use of the motors' power
		local throttle_max = max(throttle_result_FL, throttle_result_FR, throttle_result_RL, throttle_result_RR)
		local throttle_min = min(throttle_result_FL, throttle_result_FR, throttle_result_RL, throttle_result_RR)
		
		local max_absolute_throttle = max(throttle_max, abs(throttle_min))
		
		if max_absolute_throttle > 1 then
			throttle_result_RL = throttle_result_RL / max_absolute_throttle
			throttle_result_RR = throttle_result_RR / max_absolute_throttle
			throttle_result_FL = throttle_result_FL / max_absolute_throttle
		    throttle_result_FR = throttle_result_FR / max_absolute_throttle
		end
		
		-- reverse motors if their throttle values are negative
		-- set all negative throttle values to positive after reversing motors
		if throttle_result_RL < 0 then
			motor_RL.motorDirection = -1
			throttle_result_RL = -throttle_result_RL
		else
			motor_RL.motorDirection = 1
		end
		
		if throttle_result_RR < 0 then
			motor_RR.motorDirection = -1
			throttle_result_RR = -throttle_result_RR
		else
			motor_RR.motorDirection = 1
		end
		
		if throttle_result_FL < 0 then
			motor_FL.motorDirection = -1
			throttle_result_FL = -throttle_result_FL
		else
			motor_FL.motorDirection = 1
		end
		
		if throttle_result_FR < 0 then
			motor_FR.motorDirection = -1
			throttle_result_FR = -throttle_result_FR
		else
			motor_FR.motorDirection = 1
		end
		
		-- assign throttle results and enjoy
		electrics.values.throttle_RL = throttle_result_RL
		electrics.values.throttle_RR = throttle_result_RR
		electrics.values.throttle_FL = throttle_result_FL
		electrics.values.throttle_FR = throttle_result_FR
	
	end
	
end

local function calculateCharacteristicSpeed()
	local eg = (mass * (skewStiffnessRear * distanceCOGRearAxle - skewStiffnessFront * distanceCOGFrontAxle)) / (skewStiffnessFront * skewStiffnessRear * wheelBase)
	local characteristicSpeed = sqrt(wheelBase / abs(eg + 1e-30)) --guard against infinity
	return characteristicSpeed
end


local function init(jbeamData)
  configData = jbeamData
end

local function initSecondStage()
	
	-- grab all the wheels and cache them
	for _,wd in pairs(wheels.wheels) do
		if (wd.name ==  "FL") then
			wheel_FL = wd
		elseif (wd.name == "FR") then
			wheel_FR = wd
		elseif (wd.name == "RL") then
			wheel_RL = wd
		elseif (wd.name == "RR") then
			wheel_RR = wd
		end
	end
	
	wheelBase = (vec3(v.data.nodes[wheel_FL.node1].pos) - vec3(v.data.nodes[wheel_RL.node1].pos)):length()
	
	invWheelBase = 1/wheelBase
	
	yawSmooth = newExponentialSmoothing(50)
	
	local tmp = vec3(0, 0, 0)
	local totalMass = 0
	for _, v in pairs(v.data.nodes) do
		tmp = tmp + vec3(v.pos) * v.nodeWeight
		totalMass = totalMass + v.nodeWeight
	end

	local realCOG = tmp / totalMass
	
	mass = totalMass
	
	--Find the positions of the front and rear axle
	local frontAxlePos = 0
	local rearAxlePos = 0
	local twLeft = 0
	local twRight = 0
	for _, n in pairs(v.data.nodes) do
		if n.cid == wheel_FR.node1 then
			frontAxlePos = n.pos.y
			twRight = n.pos.x
		elseif n.cid == wheel_RR.node1 then
			rearAxlePos = n.pos.y
		elseif n.cid == wheel_FR.node1 then
			twLeft = n.pos.x
		end
	end
	
	
	
	distanceCOGFrontAxle = abs(realCOG.y - frontAxlePos)
	distanceCOGRearAxle = abs(realCOG.y - rearAxlePos)
	
	characteristicSpeed = calculateCharacteristicSpeed()
	
	invSquaredCharacteristicSpeed = 1 / (characteristicSpeed * characteristicSpeed)
	
	M.updateWheelsIntermediate = updateWheelsIntermediate
	M.updateGFX = updateGFX
	
end


-- public interface
M.init = init
M.initSecondStage = initSecondStage
M.updateWheelsIntermediate = nil
M.updateGFX = nil
M.update = update

return M